(**********************************************************************

    :Program.    PathFinder.mod
    :Contents.   m2Amiga project path handler for Make [bne]
    :Author.     Nicolas Benezan [bne]
    :Address.    Postwiesenstr. 2, D7000 Stuttgart 60
    :Phone.      711/333679
    :Copyright.  Shareware, see Doc-File
    :Language.   Modula-2
    :Translator. M2Amiga A+L V3.2d
    :Imports.    Environment1.2, MemSystem1.3, ErrorReq [bne]
    :Imports.    AVL1.2, List1.1 [mif]
    :History.    V1.0 [bne] 7.Feb.1989
    :History.    V1.1 [bne] 10.Feb.1989 (+ FindLocal)
    :History.    V1.2 [bne] 14.Feb.1989 (bugs fixed)
    :History.    V1.3 [bne] 15.Feb.1989 (+ DirLock)
    :History.    V1.4d [bne] 22.Feb.1989 (+ Assert(globalPathLoaded))
    :Update.     4.Mar.1989 [bne] adapted MemSystem V1.3
    :History.    V1.5 [bne] 4.Apr.1989 (+ BinDir, cosmetics)
    :History.    V1.6 [bne] 18.May.1989 (+ ExDos)
    :History.    V1.7 [bne] 13.Aug.1989 (- AVL, + Trees)

**********************************************************************)

IMPLEMENTATION MODULE PathFinder;

FROM Arts        IMPORT ModName;
FROM ExDos       IMPORT DupLock, Examine, FileInfoBlock, FileLockPtr,
                        GetCurrentDir, Lock, ParentDir, SetCurrentDir,
                        sharedLock, UnLock;
FROM Dos         IMPORT Close, FileHandlePtr, oldFile, Open, Read;
FROM MemSystem   IMPORT Deallocate, NoCareAllocate;
FROM Environment IMPORT MainDir, SystemDir;
FROM ErrorReq    IMPORT Assert;
FROM Strings     IMPORT Compare, Insert;
FROM Str         IMPORT Concat, Length;
FROM SYSTEM      IMPORT ADDRESS, ADR, BPTR, CAST, SETREG;
FROM Trees       IMPORT DiscardTree, Do, Get, InitTree, Put, Tree;
IMPORT ASCII, List;

        (* Error messages *)
CONST
  ProjNotFound    = "ProjectLock not found";
  AddProjFail     = "AddProject() failed";
  ListInitFail    = "List.Init(lockList) failed";
  NoPathFile      = "unlikely to be a path file";
  PathNotOpened   = "can't Open(PathFile)";
  PathNotFound    = "path not found";
  InitTreeFail    = "InitTree() failed";
  NoGlobalPath    = "global path not found";

TYPE
  ProjectLock=FileLockPtr;
  PathListStatus=(unknown, loaded, examined);
  CharPtr=POINTER TO CHAR;
  PathList=RECORD
    lockList: List.list;
    bufferPos: CharPtr;
    bufferEnd: CharPtr;
    status: PathListStatus;
  END;
  ProjectInfo=RECORD
    lock: FileLockPtr;                       (* project dir *)
    flags: PathFlagSet;
    subDir: ARRAY [sym..imp] OF FileLockPtr; (* sym,obj,txt dir *)
    localPath: PathList;
  END;
  ProjectInfoPtr=POINTER TO ProjectInfo;

VAR
  LockTree: Tree;
  GlobalPath: PathList;
  CurrentProject, MainLock: ProjectLock;
  CurrentInfo: ProjectInfoPtr;

(**)
(* Compare procedure for Tree                                           *)
(**)
PROCEDURE ComparePointers(Adr1, Adr2: ADDRESS): INTEGER;
  VAR
    Ptr1, Ptr2: POINTER TO LONGINT;
  BEGIN
    Ptr1:=Adr1;
    Ptr2:=Adr2;
    IF Ptr1^<Ptr2^ THEN
      RETURN -1;
    ELSIF Ptr1^=Ptr2^ THEN
      RETURN 0;
    ELSE
      RETURN 1;
    END;
  END ComparePointers;

(**)
(* Full path name of the current directory                              *)
(**)
PROCEDURE GetCurrentPath(VAR Path:ARRAY OF CHAR);
  VAR
    Info:FileInfoBlock;
    Pos:INTEGER;
    Parent,Dir:FileLockPtr;
  BEGIN
    Path[0]:=CHR(0);
    Pos:=0;
    Dir:=DupLock(GetCurrentDir());
    WHILE Dir#NIL DO
      IF Examine(Dir,Info) AND (Info.dirEntryType>0) THEN
        Insert(Path,0,"/");
        Insert(Path,0,Info.fileName);
        Pos:=Length(Info.fileName);
      END;
      Parent:=ParentDir(Dir);
      UnLock(Dir);
      Dir:=Parent;
    END;
    IF Pos#0 THEN
      Path[Pos]:=":";
    END;
  END GetCurrentPath;

PROCEDURE DirLock(Project: ProjectLock;
                  FileType: PathFlags): FileLockPtr;
  BEGIN
    IF Project#CurrentProject THEN
      CurrentInfo:=Get(LockTree, ADR(Project));
      Assert(CurrentInfo#NIL, ADR(ProjNotFound));
      CurrentProject:=Project;
    END;
    WITH CurrentInfo^ DO
      IF FileType IN flags THEN
        RETURN subDir[FileType];
      ELSE
        RETURN BPTR(Project);
      END;
    END;
  END DirLock;

(**)
(* Set current dir to a project (and probably sym/obj/txt subdirs)      *)
(**)
PROCEDURE GoToProject(Project:ProjectLock;FileType:PathFlags);
  BEGIN
    SetCurrentDir(DirLock(Project,FileType));
  END GoToProject;

(**)
(* Examine a project dir and add it to the Tree                         *)
(**)
PROCEDURE AddProject(NewProjectDir: FileLockPtr): ProjectLock;
  VAR
    Info: ProjectInfo;
    SubDir: FileLockPtr;
  BEGIN
    SetCurrentDir(NewProjectDir);
    WITH Info DO
      lock:=NewProjectDir;
      SubDir:=Lock(DefDir, sharedLock); (* "txt" *)
      IF SubDir#NIL THEN
        subDir[def]:=SubDir;
        subDir[imp]:=SubDir;
        flags:=PathFlagSet{def, imp};
      ELSE
        flags:=PathFlagSet{};
      END;
      SubDir:=Lock(SymDir, sharedLock); (* "sym" *)
      IF SubDir#NIL THEN
        subDir[sym]:=SubDir;
        INCL(flags, sym);
      END;
      SubDir:=Lock(ObjDir, sharedLock); (* "obj" *)
      IF SubDir#NIL THEN
        subDir[obj]:=SubDir;
        INCL(flags, obj);
      END;
      localPath.status:=unknown; (* path file has not been examined yet *)
    END; (* with Info *)
    Assert(Put(LockTree, Info)#NIL, ADR(AddProjFail));
    RETURN CAST(ProjectLock, Info.lock)
  END AddProject;

(**)
(* Load a local path file (if found)                                    *)
(**)
PROCEDURE LoadPathFile(VAR list:PathList);
  VAR
    PathFileLock:FileLockPtr;
    PathFile:FileHandlePtr;
    InfoBlock:FileInfoBlock;
  BEGIN
    WITH list DO
      status:=examined;
      Assert(List.Init(lockList),ADR(ListInitFail));
      PathFileLock:=Lock(PathFileName,sharedLock);
      IF PathFileLock#NIL THEN
        IF Examine(PathFileLock,InfoBlock) THEN
          UnLock(PathFileLock);
          WITH InfoBlock DO
            Assert((size>1) AND (size<2000),ADR(NoPathFile));
            NoCareAllocate(bufferPos,size);
            PathFile:=Open(ADR(PathFileName),oldFile);
            Assert(PathFile#NIL,ADR(PathNotOpened));
            IF Read(PathFile,bufferPos,size)=size THEN
              status:=loaded;
              bufferEnd:=CharPtr(LONGINT(bufferPos)+size);
            END;
            Close(PathFile);
          END; (* with Info Block *)
        ELSE
          UnLock(PathFileLock);
        END; (* if Examine() *)
      END; (* if PathFileLock#NIL *)
    END; (* with list *)
  END LoadPathFile;

(**)
(* scan path file, mark end of line with ASCII.nul                      *)
(**)
PROCEDURE ScanBuffer(VAR Pos:CharPtr;End:CharPtr):ADDRESS;
  VAR
    StartPos:CharPtr;
  BEGIN
    StartPos:=Pos;
    DEC(End);
    WHILE (Pos^>=" ") AND (LONGINT(Pos)<LONGINT(End)) DO
      INC(Pos);       (* skip pathname *)
    END;
    INC(End);
    Pos^:=ASCII.nul;
    WHILE (Pos^<" ") AND (LONGINT(Pos)<LONGINT(End)) DO
      INC(Pos);       (* skip lf, cr *)
    END;
    RETURN ADDRESS(StartPos);
  END ScanBuffer;

(**)
(* Append an extension to a file name                                   *)
(**)
PROCEDURE ExtendFileName(VAR Name:ARRAY OF CHAR;Type:PathFlags);
  VAR
    Extens:ARRAY [0..4] OF CHAR;
  BEGIN
    CASE Type OF
      |def: Extens:=DefExtens;
      |imp: Extens:=ModExtens;
      |sym: Extens:=SymExtens;
      |obj: Extens:=ObjExtens;
    ELSE
      RETURN (* FileType=proj *)
    END;
    Concat(Name,Extens);
  END ExtendFileName;

(**)
(* Search a module tracing the paths (first local then global)          *)
(**)
PROCEDURE FindProject(VAR CurrentProj:ProjectLock;
                          ModuleName:ModName;
                          FileType:PathFlags):FileLockPtr;
VAR
  File:FileLockPtr;
  Info:ProjectInfoPtr;

  PROCEDURE SearchPath(VAR list:PathList);
    VAR
      NewLock:FileLockPtr;
      NewPath:POINTER TO ARRAY [0..80] OF CHAR;
    BEGIN
      WITH list DO
        List.Reset(lockList);
        LOOP
          IF List.End(lockList) THEN
            IF status=loaded THEN
              NewPath:=ScanBuffer(bufferPos,bufferEnd);
              NewLock:=Lock(NewPath^,sharedLock);
              Assert(NewLock#NIL,ADR(PathNotFound));
              CurrentProj:=AddProject(NewLock);
              IF LONGINT(bufferPos)>=LONGINT(bufferEnd) THEN
                status:=examined;
              END;
              List.Write(lockList,CurrentProj);
            ELSE
              EXIT
            END;
          ELSE
            List.Read(lockList,CurrentProj);
          END;
          GoToProject(CurrentProj,FileType);
          File:=Lock(ModuleName,sharedLock);
          IF File#NIL THEN
            EXIT
          END;
        END;
      END;
    END SearchPath;

  BEGIN
    ExtendFileName(ModuleName,FileType);
    GoToProject(CurrentProj,FileType);    (* set CurrentDir *)
    File:=Lock(ModuleName,sharedLock);
    IF File=NIL THEN
      Info:=CurrentInfo;
      IF Info^.localPath.status=unknown THEN
        GoToProject(CurrentProj,proj);
        LoadPathFile(Info^.localPath);
      END;
      SearchPath(Info^.localPath);
      IF File=NIL THEN
        SearchPath(GlobalPath);
      END;
    END;
    RETURN File;
  END FindProject;

(**)
(* Find a file in a project drawer or sym/obj/txt subdirs               *)
(**)
PROCEDURE FindLocal(CurrentProj:ProjectLock;
                    ModuleName:ModName;
                    FileType:PathFlags):FileLockPtr;
  BEGIN
    GoToProject(CurrentProj,FileType);
    ExtendFileName(ModuleName,FileType);
    RETURN Lock(ModuleName,sharedLock);
  END FindLocal;

(**)
(* Root of path searching                                               *)
(**)
PROCEDURE MainProject():ProjectLock;
  BEGIN
    RETURN MainLock;
  END MainProject;

(**)
(* Cleanup the Tree                                                     *)
(**)
PROCEDURE FreeLocks(InfoAdr, Unused: ADDRESS);
  VAR
    SubFlag: PathFlags;
    InfoPtr: ProjectInfoPtr;
  BEGIN
    InfoPtr:=InfoAdr;
    WITH InfoPtr^ DO
      FOR SubFlag:=sym TO def DO (* sym, obj, txt *)
        IF SubFlag IN flags THEN
          UnLock(subDir[SubFlag]);
        END;
      END;
      IF localPath.status#unknown THEN
        List.Discard(localPath.lockList);
      END;
      UnLock(lock);              (* proj *)
    END;
  END FreeLocks;

PROCEDURE FreeProjects;
  BEGIN
    SetCurrentDir(SystemDir); (* avoid UnLock(CurrentDir) *)
    Do(LockTree, FreeLocks, NIL);
    DiscardTree(LockTree);
  END FreeProjects;

(**)
(* Examine the main project, load the global path                       *)
(**)
PROCEDURE InitGlobal;
  VAR
    MainPath:ARRAY [0..80] OF CHAR;
    Len:INTEGER;
  BEGIN
    Assert(InitTree(LockTree, FALSE, ComparePointers), ADR(InitTreeFail));
    SetCurrentDir(MainDir);
    GetCurrentPath(MainPath);
    Len:=Length(MainPath);
    IF (Len>4) AND (Compare(MainPath, Len-4, 3, ModDir, FALSE)=0) THEN
      MainLock:=AddProject(ParentDir(MainDir));(* project/txt/MainModule *)
    ELSE
      MainLock:=AddProject(DupLock(MainDir));  (* project/MainModule *)
    END;
    SetCurrentDir(SystemDir);
    LoadPathFile(GlobalPath);
    Assert(GlobalPath.status=loaded, ADR(NoGlobalPath));
  END InitGlobal;

BEGIN
  MainLock:=NIL;
  CurrentProject:=NIL;
END PathFinder.

