(* ------------------------------------------------------------------------
  :Program.       DosHelp
  :Contents.      some *ever* needed procedures
  :Author.        Kai Bolay [kai]
  :Address.       Hoffmannstrae 168
  :Address.       7250 Leonberg 1
  :Phone.         07152/22135
  :History.       v1.0 [kai] 26-Jun-90 splitted from DiskyFuncs
  :Copyright.     PD
  :Language.      Oberon
  :Translator.    OBERON v1.08
------------------------------------------------------------------------ *)
MODULE DosHelp;

IMPORT d   : Dos,
       e   : Exec,
       so  : StringOps,
       sys : SYSTEM;

PROCEDURE SplitPath * (Path : ARRAY OF CHAR; VAR Dir, File : ARRAY OF CHAR);
(* ------------------------------------------------------------------------
  :Input.    Path: kompletter Pfad
  :Output.   Dir, File : Zerlegte Bestandteile des Pfads
  :Semantic. Zerlegt Pfad in seine Bestandteile
------------------------------------------------------------------------ *)

VAR Pos, Len : INTEGER;

BEGIN
   Len := so.Length (Path);
   Pos := so.FindChar (Path, '/', -Len+1);
   IF Pos = -1 THEN
      Pos := so.FindChar (Path, ':', -Len+1);
      IF Pos = -1 THEN
         so.Assign (Path, File);
         Dir[0] := 0X;
         RETURN;
      END; (* IF *)
   END; (* IF *)
   so.SubString (Path, File, Pos+1, so.Length (Path)-Pos-1);
   so.SubString (Path, Dir, 0, Pos+1);
END SplitPath;


PROCEDURE MergePath * (Dir, File : ARRAY OF CHAR; VAR Path : ARRAY OF CHAR);
(* ------------------------------------------------------------------------
  :Input.    Dir, File: Dir und File bzw. SubDir
  :Output.   Path: Zusammengesetzter Pfad
  :Semantic. Setzt Dir und File bzw. SubDir zusammen
------------------------------------------------------------------------ *)

VAR Last : INTEGER;

BEGIN
   so.Assign (Dir, Path);
   IF Path[0] # 0X THEN
      Last := so.Length (Path)-1;
      IF (Path[Last] # ':') AND (Path[Last] # '/') THEN
         so.AppendChar (Path, '/');
      END; (* IF *)
   END; (* IF *)
   so.Append (Path, File);
END MergePath;


PROCEDURE GetPathFromLock * (VAR Path : ARRAY OF CHAR; TheLock : d.FileLockPtr);
(* ------------------------------------------------------------------------
  :Input.    TheLock = Ein gltiger FileLock auf Directory oder File
  :Output.   Path : Ausgeschriebener Pfadname
  :Semantic. Generiert Pfadname aus Lock
------------------------------------------------------------------------ *)

VAR CurDirPtr : d.FileLockPtr;
    OldDirPtr : d.FileLockPtr;
    FIBPtr    : d.FileInfoBlockPtr;
    VolumeLen : INTEGER;

BEGIN
   Path[0] := 0X;
   CurDirPtr := d.DupLock (TheLock);
   IF CurDirPtr = NIL THEN RETURN END;
   e.Forbid;
   so.SubString (CurDirPtr^.volume^.name^, Path, 1,
                  sys.VAL (SHORTINT, CurDirPtr^.volume^.name^[0]));
   e.Permit;
   so.AppendChar (Path, ':');
   VolumeLen := so.Length (Path);
   FIBPtr := e.AllocMem (sys.SIZE (FIBPtr^), LONGSET {});
   IF FIBPtr # NIL THEN
      WHILE CurDirPtr # NIL DO
         IF NOT (d.Examine (CurDirPtr, FIBPtr)) THEN
            Path[0] := 0X;
            d.UnLock (CurDirPtr);
            CurDirPtr := NIL;
         ELSE
            OldDirPtr := CurDirPtr;
            CurDirPtr := d.ParentDir (OldDirPtr);
            d.UnLock (OldDirPtr);
            IF CurDirPtr # NIL THEN
               IF so.Length (Path) # VolumeLen THEN
                  so.InsertChar (Path, '/', VolumeLen);
               END; (* IF *)
               so.InsertSubString (Path, FIBPtr^.fileName, VolumeLen);
            END; (* IF *)
         END; (* IF *)
      END; (* WHILE *)
      e.FreeMem (FIBPtr, sys.SIZE (FIBPtr^));
   END; (* IF *)
END GetPathFromLock;


PROCEDURE GetFullPath * (VAR Path : ARRAY OF CHAR) : BOOLEAN;
(* ------------------------------------------------------------------------
  :Input.    Ein existierender Pfad (z.B. "AMOK:Programme")
  :Output.   Ausgeschriebener Pfadname (z.B. "Harddisk:PD/Amok/Programme")
  :Result.   Existiert der Pfad berhaupt?
  :Semantic. Generiert Pfadname aus Lock
------------------------------------------------------------------------ *)

VAR lock : d.FileLockPtr;

BEGIN
   lock := d.Lock (sys.ADR (Path), d.accessRead);
   IF lock # NIL THEN
      GetPathFromLock (Path, lock);
      d.UnLock (lock);
      RETURN TRUE;
   ELSE
      RETURN FALSE;
   END; (* IF *)
END GetFullPath;


PROCEDURE HelpExists (Path : ARRAY OF CHAR; What : SHORTINT) : BOOLEAN;

VAR TestLockPtr : d.FileLockPtr;
    TestFIB     : d.FileInfoBlockPtr;
    Result      : BOOLEAN;

BEGIN
   TestFIB := e.AllocMem (sys.SIZE (TestFIB^), LONGSET {});
   IF TestFIB # NIL THEN
      TestLockPtr := d.Lock (sys.ADR (Path), d.sharedLock);
      IF TestLockPtr # NIL THEN
         IF d.Examine (TestLockPtr, TestFIB) THEN
            IF What < 0 THEN Result := (TestFIB^.dirEntryType < 0) END;
            IF What > 0 THEN Result := (TestFIB^.dirEntryType > 0) END;
            IF What = 0 THEN Result := TRUE; END;
         ELSE
            Result := FALSE;
         END; (* IF *)
         d.UnLock (TestLockPtr);
      ELSE
         Result := FALSE;
      END; (* IF *)
      e.FreeMem (TestFIB, sys.SIZE (TestFIB^));
   ELSE
      RETURN FALSE;
   END; (* IF *)
   RETURN Result;
END HelpExists;


PROCEDURE FileExists * (File : ARRAY OF CHAR) : BOOLEAN;
(* ------------------------------------------------------------------------
  :Input.    File = Name des zu prfenden Files
  :Result.   Existiert das File?
  :Semantic. Prft ob File existiert
------------------------------------------------------------------------ *)
BEGIN
   RETURN HelpExists (File, -1);
END FileExists;


PROCEDURE DirExists * (Dir : ARRAY OF CHAR) : BOOLEAN;
(* ------------------------------------------------------------------------
  :Input.    Dir = Name des zu prfenden Dirs
  :Result.   Existiert das Dir?
  :Semantic. Prft ob Dir existiert
------------------------------------------------------------------------ *)
BEGIN
   RETURN HelpExists (Dir, +1);
END DirExists;


PROCEDURE Exists * (Path : ARRAY OF CHAR) : BOOLEAN;
(* ------------------------------------------------------------------------
  :Input.    Path = Name des zu prfenden Dirs/Files
  :Result.   Existiert der Pfad (Dir / File)?
  :Semantic. Prft ob Pfad (Dir / File) existiert
------------------------------------------------------------------------ *)
BEGIN
   RETURN HelpExists (Path, 0);
END Exists;


END DosHelp.
