(**********************************************************************

    :Program.    StringOps.mod
    :Contents.   string operations, replaces the standard module "Strings".
    :Author.     Nicolas Benezan [bne]
    :Address.    Postwiesenstr. 2, D7000 Stuttgart 60
    :Phone.      711/333679
    :Copyright.  Public Domain
    :Language.   Oberon
    :Translator. OBERON v1.01
    :History.    V1.0 [bne] 18-Apr-89
    :History.    V1.1 [bne] 18-Apr-89 (bug in DeleteSubString fixed)
    :History.    V2.0 [kai] 25-Apr-90 ported to Oberon

**********************************************************************)

MODULE StringOps;

(* All procedures will truncate the string if it was too long to
   fit into the array boundaries. *)

CONST
  nul=00X;

PROCEDURE Length * (String:ARRAY OF CHAR):INTEGER;
(*:Result.      Length of <String>
*)
  VAR
    Len:INTEGER;
  BEGIN
    Len:=0;
    WHILE (Len<=LEN(String)) AND (String[Len]#nul) DO
      INC(Len);
    END;
    RETURN Len;
  END Length;


PROCEDURE Append * (VAR String:ARRAY OF CHAR;
                     Tail:ARRAY OF CHAR);
(*:Semantic.    Appends <Tail> to <String>
*)
  VAR
    PosStr,PosTail:INTEGER;
  BEGIN
    PosTail:=0;
    PosStr:=Length(String);
    LOOP
      IF PosStr>LEN(String) THEN
        EXIT
      ELSIF (PosTail>LEN(Tail)) OR (Tail[PosTail]=nul) THEN
        String[PosStr]:=nul;
        EXIT
      END;
      String[PosStr]:=Tail[PosTail];
      INC(PosStr);
      INC(PosTail);
    END;
  END Append;

PROCEDURE AppendChar * (VAR String:ARRAY OF CHAR;
                         Char:CHAR);
(*:Semantic.    Appends a single character to <String>
*)
  VAR
    Len:INTEGER;
  BEGIN
    Len:=Length(String);
    IF Len<=LEN(String) THEN
      String[Len]:=Char;
      IF Len<LEN(String) THEN
        String[Len+1]:=nul;
      END;
    END;
  END AppendChar;

PROCEDURE Assign * (    Source:ARRAY OF CHAR;
                 VAR Destination:ARRAY OF CHAR);
(*:Semantic.    Copies <Source> to <Destination>.
*)
  VAR
    Pos,Max:INTEGER;
  BEGIN
    Max:=LEN(Source);
    IF LEN(Destination)<Max THEN
      Max:=LEN(Destination);
    END;
    Pos:=0;
    LOOP
      Destination[Pos]:=Source[Pos];
      INC(Pos);
      IF (Pos>Max) OR (Source[Pos]=nul) THEN
        IF Pos<=LEN(Destination) THEN
          Destination[Pos]:=nul;
        END;
        EXIT
      END;
    END;
  END Assign;

PROCEDURE CapChar * (VAR Char:CHAR);
(*:Semantic.    Returns the capital letter of <Char>
  :Note.        works also with foreign letters (...)
*)
  BEGIN
    CASE Char OF
     |"a".."z", "".."", "".."": DEC(Char,32);
    ELSE
    END;
  END CapChar;

PROCEDURE CapString * (VAR String:ARRAY OF CHAR);
(*:Semantic.    CapChar()s each character of <String>
*)
  VAR
    Pos:INTEGER;
  BEGIN
    Pos:=0;
    WHILE (Pos<=LEN(String)) AND (String[Pos]#nul) DO
      CapChar(String[Pos]);
      INC(Pos);
    END;
  END CapString;

PROCEDURE FindChar * (String:ARRAY OF CHAR;
                   Char:CHAR;
                   Start:INTEGER):INTEGER;
(*:Semantic.    Searches next occurence of <Char> in <String>
  :Input.       Start: position where to start searching
  :Result.      position of <Char> in <String>
  :Result.      or -1 if <Char> is not found
*)
  VAR
    Step:INTEGER;
  BEGIN
    IF Start<0 THEN
      Start:=-Start;
      Step:=-1;
    ELSE
      Step:=1;
    END;
    WHILE (Start<=LEN(String)) AND (Start>=0) AND (String[Start]#nul) DO
      IF String[Start]=Char THEN
        RETURN Start
      END;
      INC(Start,Step);
    END;
    RETURN -1;
  END FindChar;

PROCEDURE Concat * (    Head:ARRAY OF CHAR;
                     Tail:ARRAY OF CHAR;
                 VAR String:ARRAY OF CHAR);
(*:Semantic.    concatenates <Head> and <Tail> to <String>
*)
  BEGIN
    Assign(Head,String);
    Append(String,Tail);
  END Concat;

PROCEDURE DeleteSubString * (VAR String:ARRAY OF CHAR;
                              Start,Len:INTEGER);
(*:Semantic.    Deletes a substring of <String> which starts
  :Semantic.    at position <Start> and is <Len> characters
  :Semantic.    long.
*)
  BEGIN
    INC(Len,Start);
    IF Len >= Length (String) THEN
      String[Start]:= nul;
      RETURN
    END;
    WHILE Len<=LEN(String) DO
      String[Start]:=String[Len];
      IF String[Len]=nul THEN
        RETURN
      END;
      INC(Start);
      INC(Len);
    END;
    IF Start<=LEN(String) THEN
      String[Start]:=nul;
    END;
  END DeleteSubString;

PROCEDURE FindSubString * (String:ARRAY OF CHAR;
                        SubString:ARRAY OF CHAR;
                        Start:INTEGER):INTEGER;
(*:Semantic.    Searches next occurence of <SubString> in <String>
  :Input.       Start: position where to start searching
  :Result.      next position of <SubString> in <String>
  :Result.      or -1 if not found
*)
  VAR
    Pos,Max:INTEGER;
  BEGIN
    Pos:=0;
    Max:=LEN(String)-Length(SubString)+1;
    WHILE (String[Start]#nul) AND (Start<=Max) DO
      LOOP
        IF (Pos>LEN(SubString)) OR (SubString[Pos]=nul) THEN
          RETURN Start
        END;
        IF String[Start+Pos]#SubString[Pos] THEN
          Pos:=0;
          EXIT
        END;
        INC(Pos);
      END;
      INC(Start);
    END;
    RETURN -1;
  END FindSubString;

PROCEDURE ShiftRight (VAR String:ARRAY OF CHAR;
                         Position, Distance: INTEGER);
  VAR
    Start,End:INTEGER;
  BEGIN
    End:=Length(String)+Distance;
    IF End>LEN(String) THEN
      End:=LEN(String);
    END;
    Start:=End-Distance;
    WHILE Start>=Position DO
      String[End]:=String[Start];
      DEC(End); DEC (Start);
    END;
  END ShiftRight;

  PROCEDURE OverWrite * (VAR String:ARRAY OF CHAR;
                        Overlay:ARRAY OF CHAR;
                        Position:INTEGER);
(*:Semantic.    Replaces (overwrites) the caracters of <String> with
  :Semantic.    <Overlay> starting at <Position>.
*)
  VAR
    Pos,Len:INTEGER;
  BEGIN
    Len:=Length(Overlay);
    DEC(Len);
    IF Position+Len>LEN(String) THEN
      Len:=LEN(String)-Position;
    END;
    Pos := 0;
    WHILE Pos<=Len DO
      String[Position]:=Overlay[Pos];
      INC(Position); INC (Pos);
    END;
  END OverWrite;

PROCEDURE InsertSubString * (VAR String:ARRAY OF CHAR;
                              SubString:ARRAY OF CHAR;
                              Position:INTEGER);
(*:Semantic.    Inserts <SubString> into <String> at <Position>
*)
  BEGIN
    ShiftRight(String,Position,Length(SubString));
    OverWrite(String,SubString,Position);
  END InsertSubString;

PROCEDURE InsertChar * (VAR String:ARRAY OF CHAR;
                         Char:CHAR;
                         Position:INTEGER);
(*:Semantic.    Inserts a single character into <String> at <Position>
*)
  BEGIN
    ShiftRight(String,Position,1);
    String[Position]:=Char;
  END InsertChar;

PROCEDURE SubString * (    Source:ARRAY OF CHAR;
                    VAR Destination:ARRAY OF CHAR;
                        Position,Len:INTEGER);
(*:Output.      Destination: the substring of <String> which starts
  :Semantic.    at position <Start> and is <Len> characters long.
*)
  VAR
    SrcLen,Pos:INTEGER;
  BEGIN
    SrcLen:=Length(Source);
    IF Position+Len>SrcLen THEN
      Len:=SrcLen-Position;
    END;
    Pos := 0;
    WHILE Pos<=Len-1 DO
      Destination[Pos]:=Source[Position];
      INC(Position); INC (Pos);
    END;
    IF Len<=LEN(Destination) THEN
      Destination[Len]:=nul;
    END;
  END SubString;

PROCEDURE Compare * (String1,String2:ARRAY OF CHAR):INTEGER;
(*:Semantic.    Compares <String1> vs <String2>
  :Result.      0 if they are equal
  :Result.      <0 if String1 < String2
  :Result.      >0 if String1 > String2
*)
  VAR
    Max,Max1,Max2,Res,Pos,Diff:INTEGER;
  BEGIN
    Max1:=LEN(String1);
    Max2:=LEN(String2);
    Res:=Max1-Max2;
    IF Res<0 THEN
      Max:=Max1;
    ELSE
      Max:=Max2;
    END;
    Pos := 0;
    WHILE Pos<=Max DO
      Diff:=ORD(String1[Pos])-ORD(String2[Pos]);
      IF (Diff#0) OR (String1[Pos]=nul) THEN
        RETURN Diff
      END;
      INC (Pos);
    END;
    INC(Max);
    IF (Res<0) AND (String2[Max]#nul) THEN
    ELSIF (Res>0) AND (String1[Max]#nul) THEN
    ELSE
      RETURN 0;
    END;
    RETURN Res;
  END Compare;

END StringOps.
