(*************************************************************************

:Program.    MemPool.mod
:Contents.   Block oriented memory management
:Author.     Hartmut Goebel [hG]
:Address.    Aufseplatz 5, D-8500 Nrnberg 40
:Address.    UseNet: hartmut@oberon.nbg.sub.org
:Address.    Z-Netz: hartmut@asn.zer   Fido: 2:246/81.1
:Copyright.  Copyright  1992 by Hartmut Goebel
:Language.   Oberon
:Translator. Amiga Oberon V2.25
:History.    V1.0, 28 May 1992 [hG]
:Date.       30 May 1992 18:52:22

*************************************************************************)

MODULE MemPool;

IMPORT
  e  := Exec,
  es := ExecSupport,
  s  := SYSTEM,
  ol := OberonLib;

TYPE
  MemHeaderPtr = UNTRACED POINTER TO MemHeader;
  MemHeader = STRUCT (mh: e.MemHeader)
    size: LONGINT;
  END;

  MemElementPtr = UNTRACED POINTER TO MemElement;
  MemElement = STRUCT
    size: LONGINT;
    mem: INTEGER;  (* only a dummy *)
  END;

VAR
  AllocBlockSize * : LONGINT;
  MemList : e.List;
  mh: MemHeaderPtr;

(* $StackChk- $RangeChk- $NilChk- $OvflChk- $CaseChk- $ReturnChk- *)

PROCEDURE ClearMem(adr{8+0}: e.APTR; size{0}: LONGINT); (* $EntryExitCode- *)
BEGIN
s.INLINE(02E00H,       (*    move.l  d0,d7      *)
         07C07H,       (*    moveq   #7,d6      *)
         0CE86H,       (*    and.l   d6,d7      *)
         06706H,       (*    beq     l          *)
         0E680H,       (*    asr.l   #3,d0      *)
         05280H,       (*    addq.l  #1,d0      *)
         0E780H,       (*    asl.l   #3,d0      *)
         05180H,       (* l: subq.l  #8,d0      *)
         042B0H,0800H, (*    clr.l   0(a0,d0.l) *)
         042B0H,0804H, (*    clr.l   4(a0,d0.l) *)
         04A80H,       (*    tst.l   d0         *)
         066F2H,       (*    bne     l          *)
         04E75H);      (*    rts                *)
END ClearMem;


PROCEDURE AllocNewBlock(size: LONGINT; requirements: LONGSET): MemHeaderPtr;
VAR
  mc: e.MemChunkPtr;
  mh: MemHeaderPtr;
BEGIN
  IF (size MOD e.blockSize) # 0 THEN
    size := ((size DIV e.blockSize) + 1) * e.blockSize; END;
  mh := e.AllocMem(SIZE(MemHeader),LONGSET{});
  IF mh = NIL THEN RETURN NIL; END;
  mc := e.AllocMem(size,requirements - LONGSET{e.memClear});
  IF mc = NIL THEN e.FreeMem(mh,SIZE(MemHeader)); RETURN NIL; END;

  mh.mh.node.type := e.memory;
  mh.mh.node.name := NIL;
  mh.mh.first := mc;
  mh.mh.lower := mc;
  mh.mh.upper := s.VAL(s.ADDRESS,s.VAL(LONGINT,mc)+size);
  mh.mh.free := size;
  (* mh.mh.attributes := requirements * SET{0..15}; *)
  mh.size := size;

  mc.next := NIL;
  mc.bytes := size;

  e.AddHead(MemList,mh);
  RETURN mh;
END AllocNewBlock;


PROCEDURE AllocMem * (size: LONGINT; requirements: LONGSET): e.APTR;
VAR
  mh: MemHeaderPtr;
  mem: e.APTR;
BEGIN
  IF size = 0 THEN RETURN NIL; END;
  IF size > AllocBlockSize THEN
    mh := AllocNewBlock(size,requirements);
  ELSE
    mh := MemList.head(MemHeader);
    WHILE mh.mh.node.succ # NIL DO
      mem := e.Allocate(mh,size);
      IF mem # NIL THEN
        IF e.memClear IN requirements THEN ClearMem(mem,size); END;
        RETURN mem;
      END;
      mh := mh.mh.node.succ(MemHeader);
    END;
    mh := AllocNewBlock(AllocBlockSize,requirements);
  END;
  IF mh = NIL THEN
    RETURN NIL; END;
  mem := e.Allocate(mh,size); (* asure: mem # NIL *)
  IF e.memClear IN requirements THEN
    ClearMem(mem,size); END;
  RETURN mem;
END AllocMem;


PROCEDURE New*(VAR adr: e.APTR; size: LONGINT);
VAR
  mem: MemElementPtr;
BEGIN
  INC(size,SIZE(LONGINT));
  mem := AllocMem(size,ol.MemReqs);
  IF mem # NIL THEN
    mem.size := size;
    adr := s.ADR(mem.mem);
  ELSE
    adr := NIL;
  END;
END New;


PROCEDURE FreeMem * (mem: e.APTR; size: LONGINT);
VAR
  mh: MemHeaderPtr;
CONST
  TRAP3 = 4E43H;
BEGIN
  mh := MemList.head(MemHeader);
  WHILE mh.mh.node.succ # NIL DO
    IF (s.VAL(LONGINT,mh.mh.lower) <= s.VAL(LONGINT,mem))
     & (s.VAL(LONGINT,mh.mh.upper) >  s.VAL(LONGINT,mem)) THEN
      e.Deallocate(mh,mem,size);
      mem := NIL;
      IF mh.mh.free = mh.size THEN
        e.Remove(mh);
        e.FreeMem(mh.mh.lower,mh.size);
        e.FreeMem(mh,SIZE(MemHeader));
      END;
      RETURN;
    END;
    mh := mh.mh.node.succ(MemHeader);
  END;
  s.INLINE(TRAP3);
END FreeMem;


PROCEDURE Dispose*(VAR adr: e.APTR);
VAR
  mem: MemElementPtr;
BEGIN
  IF adr # NIL THEN
    mem := s.VAL(e.APTR,s.VAL(LONGINT,adr)-SIZE(LONGINT));
    FreeMem(mem,mem.size);
    adr := NIL;
  END;
END Dispose;


BEGIN
  AllocBlockSize := 1024;
  es.NewList(MemList);

CLOSE
  mh := MemList.head(MemHeader);
  WHILE mh.mh.node.succ # NIL DO
    e.Remove(mh);
    e.FreeMem(mh.mh.lower,mh.size);
    e.FreeMem(mh,SIZE(MemHeader));
    mh := MemList.head(MemHeader);
  END;

END MemPool.

