(*************************************************************************

:Program.    MemPool.mod
:Contents.   Block oriented memory management
:Author.     Hartmut Goebel [hG]
:Address.    Aufseplatz 5, D-8500 Nrnberg 40
:Address.    UseNet: hartmut@oberon.nbg.sub.org
:Address.    Z-Netz: hartmut@asn.zer   Fido: 2:246/81.1
:Copyright.  Copyright  1992,1993 by Hartmut Goebel
:Language.   Oberon
:Translator. AmigaOberon V3.0
:History.    V1.0, 28 May 1992 [hG]
:History.    V1.1, 12 Jan 1993 [hG] Semaphore to make it Alloc/Free reentrant
:History.    V1.2, 14 Jan 1993 [hG] cares about mem-requirements (hopefully)

*************************************************************************)

MODULE MemPool;

IMPORT
  e  := Exec,
  y  := SYSTEM,
  ol := OberonLib;

TYPE
  MemHeaderPtr = UNTRACED POINTER TO MemHeader;
  MemHeader = STRUCT (mh: e.MemHeader)
    size: LONGINT;
  END;

  MemElementPtr = UNTRACED POINTER TO MemElement;
  MemElement = STRUCT
    size: LONGINT;
    mem: INTEGER;  (* only a dummy *)
  END;

VAR
  AllocBlockSize * : LONGINT;
  MemList : e.List;
  mh: MemHeaderPtr;
  memSem: e.SignalSemaphore;

(* $StackChk- $RangeChk- $NilChk- $OvflChk- $CaseChk- $ReturnChk- *)

PROCEDURE ClearMem(adr{8}: e.APTR; size{0}: LONGINT); (* $EntryExitCode- *)
BEGIN
y.INLINE(02E00H,       (*    move.l  d0,d7      *)
         07C07H,       (*    moveq   #7,d6      *)
         0CE86H,       (*    and.l   d6,d7      *)
         06706H,       (*    beq     l          *)
         0E680H,       (*    asr.l   #3,d0      *)
         05280H,       (*    addq.l  #1,d0      *)
         0E780H,       (*    asl.l   #3,d0      *)
         05180H,       (* l: subq.l  #8,d0      *)
         042B0H,0800H, (*    clr.l   0(a0,d0.l) *)
         042B0H,0804H, (*    clr.l   4(a0,d0.l) *)
         04A80H,       (*    tst.l   d0         *)
         066F2H,       (*    bne     l          *)
         04E75H);      (*    rts                *)
END ClearMem;


PROCEDURE AllocNewBlock(size: LONGINT; requirements: LONGSET): MemHeaderPtr;
TYPE
 doubleSet = ARRAY 2 OF SET;
VAR
  mc: e.MemChunkPtr;
  mh: MemHeaderPtr;
BEGIN
  IF (size MOD e.blockSize) # 0 THEN
    size := ((size DIV e.blockSize) + 1) * e.blockSize; END;
  mh := e.AllocMem(SIZE(MemHeader),LONGSET{});
  IF mh = NIL THEN RETURN NIL; END;
  mc := e.AllocMem(size,requirements - LONGSET{e.memClear});
  IF mc = NIL THEN e.FreeMem(mh,SIZE(MemHeader)); RETURN NIL; END;

  mh.mh.node.type := e.memory;
  mh.mh.node.name := NIL;
  mh.mh.first := mc;
  mh.mh.lower := mc;
  mh.mh.upper := y.VAL(y.ADDRESS,y.VAL(LONGINT,mc)+size);
  mh.mh.free := size;
  mh.mh.attributes := y.VAL(doubleSet,requirements)[1];
  mh.size := size;

  mc.next := NIL;
  mc.bytes := size;

  e.AddHead(MemList,mh);
  RETURN mh;
END AllocNewBlock;


PROCEDURE AllocMem * (size: LONGINT; requirements: LONGSET): e.APTR;
TYPE
 doubleSet = ARRAY 2 OF SET;
VAR
  mh: MemHeaderPtr;
  req: SET;
  mem: e.APTR;
BEGIN
  IF size = 0 THEN RETURN NIL; END;
  e.ObtainSemaphore(memSem);
  req := y.VAL(doubleSet,requirements)[1];
  IF size > AllocBlockSize THEN
    mh := AllocNewBlock(size,requirements);
  ELSE
    mh := MemList.head(MemHeader);
    WHILE mh.mh.node.succ # NIL DO
      IF mh.mh.attributes = req THEN
        mem := e.Allocate(mh,size);
        IF mem # NIL THEN
          IF e.memClear IN requirements THEN ClearMem(mem,size); END;
          e.ReleaseSemaphore(memSem);
          RETURN mem;
        END;
      END;
      mh := mh.mh.node.succ(MemHeader);
    END;
    mh := AllocNewBlock(AllocBlockSize,requirements);
  END;
  IF mh = NIL THEN
    e.ReleaseSemaphore(memSem);
    RETURN NIL; END;
  mem := e.Allocate(mh,size); (* asure: mem # NIL *)
  IF e.memClear IN requirements THEN
    ClearMem(mem,size); END;
  e.ReleaseSemaphore(memSem);
  RETURN mem;
END AllocMem;


PROCEDURE New*(VAR adr: e.APTR; size: LONGINT);
VAR
  mem: MemElementPtr;
BEGIN
  INC(size,SIZE(LONGINT));
  mem := AllocMem(size,ol.MemReqs);
  IF mem # NIL THEN
    mem.size := size;
    adr := y.ADR(mem.mem);
  ELSE
    adr := NIL;
  END;
END New;


PROCEDURE FreeMem * (mem: e.APTR; size: LONGINT);
VAR
  mh: MemHeaderPtr;
CONST
  TRAP3 = 4E43H;
BEGIN
  e.ObtainSemaphore(memSem);
  mh := MemList.head(MemHeader);
  WHILE mh.mh.node.succ # NIL DO
    IF (y.VAL(LONGINT,mh.mh.lower) <= y.VAL(LONGINT,mem))
     & (y.VAL(LONGINT,mh.mh.upper) >  y.VAL(LONGINT,mem)) THEN
      e.Deallocate(mh,mem,size);
      mem := NIL;
      IF mh.mh.free = mh.size THEN
        e.Remove(mh);
        e.FreeMem(mh.mh.lower,mh.size);
        e.FreeMem(mh,SIZE(MemHeader));
      END;
      e.ReleaseSemaphore(memSem);
      RETURN;
    END;
    mh := mh.mh.node.succ(MemHeader);
  END;
  e.ReleaseSemaphore(memSem);
  y.INLINE(TRAP3);
END FreeMem;


PROCEDURE Dispose*(VAR adr: e.APTR);
VAR
  mem: MemElementPtr;
BEGIN
  IF adr # NIL THEN
    mem := y.VAL(e.APTR,y.VAL(LONGINT,adr)-SIZE(LONGINT));
    FreeMem(mem,mem.size);
    adr := NIL;
  END;
END Dispose;

VAR
  initialized: INTEGER;

(*
 * This mechanism builds on the knowledge that all vars are set
 * to zero before the programm starts.
 *)

PROCEDURE Init*();
BEGIN
  IF initialized = 0 THEN
    e.InitSemaphore(memSem);
    AllocBlockSize := 1024;
    MemList.head := y.ADR(MemList.tail);
    MemList.tail := NIL;
    MemList.tailPred := y.ADR(MemList.head);
  END;
  INC(initialized);
END Init;


PROCEDURE Close*();
BEGIN
  DEC(initialized);
  IF initialized = 0 THEN
    mh := MemList.head(MemHeader);
    WHILE mh.mh.node.succ # NIL DO
      e.Remove(mh);
      e.FreeMem(mh.mh.lower,mh.size);
      e.FreeMem(mh,SIZE(MemHeader));
      mh := MemList.head(MemHeader);
    END;
  END;
END Close;

BEGIN
  Init;

CLOSE
  Close;

END MemPool.

