(***************************************************************************

:Program.    MoreStrings.mod
:Author.     Hartmut Goebel [hG]
:Copyright.  PublicDomain
:Language.   Oberon, 68000-Assembler
:Translator. Amiga Oberon V3.0
:History.    V1.0, 22 Nov 1992 [hG]
:History.    V1.1, 03 Jan 1993 [hG] -Bug in StrCmp,
:History.                           removed last INTEGER restrictions.
:Support.    some Procedures taken from N.Benezan's StringOps
:Support.    KillSpaces/ChangeBlanks based on an idea of Frank Schummertz
:Date.       03 Jan 1993 23:12:25

***************************************************************************)

MODULE MoreStrings;

IMPORT
  BT := BasicTypes,
  e  := Exec,
  str:= Strings,
  y  := SYSTEM;

(****** MoreStrings/CLength **************************************
*
*   NAME
*
*      CLength -- returns length of 'C'-style string
*
*   SYNOPSIS
*
*      CLength (s: Exec.STRPTR): LONGINT;
*
*   FUNCTION
*
*      Returns the length of a null terminated string.
*      The null charakter is not taken in accounted.
*
*      The string _must_ be null terminated.
*
*   INPUTS
*
*      str - pointer to a null terminated string
*
*   RESULT
*
*      LONGINT - length of the string
*
*   EXAMPLE
*
*      s := SYSTEM.ADR("foobar");
*      i := MoreStrings.CLength(s);
*
*   NOTES
*
*      Do not use this for Oberon strings, since they may not be null
*      terminated!!
*
*      The intention of this function is to surround problems
*      in conjunktion with Exec.STRPTR which is defined as UNTRACED
*      POINTER TO ARRAY 256 OF CHAR but may point to a longer string.
*      Strings.Length() would only notice the first 256 chars.
*
*   SEE ALSO
*
*      Strings/Length()
*
********)

PROCEDURE CLength * (s{8}: e.STRPTR): LONGINT;
(* $EntryExitCode- *)
BEGIN
  y.INLINE(
  02248H,   (* CLength move.l  a0,a1    *)
  04A18H,   (* leloop  tst.b   (a0)+    *)
  066FCH,   (*         bne     leloop   *)
  091C9H,   (*         suba.l  a1,a0    *)
  02008H,   (*         move.l  a0,d0    *)
  05380H,   (*         subq.l  #1,d0    *)
  06B02H,   (*         bmi.s   lenout ; overflow!   *)
  04E76H,   (*         trapv            *)
  04E75H    (* lenout  rts              *)
  ) (* INLINE *);
END CLength;


(****** MoreStrings/bmove ****************************************
*
*   NAME
*
*      bmove -- moves bytes
*
*   SYNOPSIS
*
*      bmove (from, to: Exec.APTR; len{0}: LONGINT);
*
*   FUNCTION
*
*      Moves <len> bytes from location pointed to by <from> to
*      localtion pointed to by <to>
*
*   INPUTS
*
*      from - pointer to source bytes
*      to   - pointer to destination
*      len  - number of bytes to be moved
*
*   NOTES
*
*      handles overlapping areas correct: if <to> is lower than
*      <from> bytes will be copied backwards.
*
*******)

PROCEDURE bmove * (from{8}, to{9}: e.APTR; len{0}: LONGINT);
(* $EntryExitCode- *)
BEGIN
  y.INLINE(
  0B3C8H,         (* bmove   cmpa.l  a0,a1       *)
  0620CH,         (*         bhi     dobm        *)
  06716H,         (*         beq     nomove      *)
  05380H,         (*         subq.l  #1,d0       *)
  012D8H,         (* bmlloop move.b  (a0)+,(a1)+ *)
  051C8H, 0FFFCH, (*         dbra.l  d0,bmlloop  *)
  04E75H,         (*         rts                 *)
  0D1C0H,         (* dobm    adda.l  d0,a0       *)
  0D3C0H,         (*         adda.l  d0,a1       *)
  05380H,         (*         subq.l  #1,d0       *)
  01320H,         (* bmloop  move.b  -(a0),-(a1) *)
  051C8H, 0FFFCH, (*         dbra.l  d0,bmloop   *)
  04E75H          (* nomove  rts                 *)
  ) (* INLINE *);
END bmove;


(****** MoreStrings/bfill ****************************************
*
*   NAME
*
*      bfill -- fills memory with a given byte
*
*   SYNOPSIS
*
*      bfill (to: Exec.APTR; fill: CHAR; num: LONGINT);
*
*   FUNCTION
*
*      fills <num> bytes starting at <start> with the byte <fill>
*
*   INPUTS
*
*      to   - pointes to area to be filled
*      fill - byte that will be filled in
*      num  - number of bytes to be filled in
*
*********)

PROCEDURE bfill * (string{8}: e.APTR; fill{0}: CHAR; num{1}: LONGINT);
(* $EntryExitCode- *)
BEGIN
  y.INLINE(
  06002H,          (* bfill   bra     testd1    *)
  010C0H,          (* loop    move.b  d0,(a0)+  *)
  051C9H, 0FFFCH,  (* testd1  dbra    d1,loop   *)
  04E75H           (*         rts               *)
  ) (* INLINE *);
END bfill;

PROCEDURE CopyCStringAdd * (from: e.STRPTR;
                         enlarge: LONGINT): BT.DynString;
VAR
  len: LONGINT;
  new: BT.DynString;
BEGIN
  IF from = NIL THEN RETURN NIL; END;
  IF enlarge < 0 THEN enlarge := 0; END;
  len := CLength(from)+1;
  y.ALLOCATE(new,len+enlarge);
  IF new=NIL THEN
    RETURN NIL;
  END;
  e.CopyMem(from^,new^,len);
  RETURN new;
END CopyCStringAdd;


PROCEDURE CopyStringAdd * (from: ARRAY OF CHAR;
                        enlarge: LONGINT): BT.DynString;
(* $CopyArrays- *)
VAR
  len: LONGINT;
  new: BT.DynString;
BEGIN
  IF enlarge < 0 THEN enlarge := 0; END;
  len := str.Length(from)+1;
  y.ALLOCATE(new,len+enlarge);
  IF new=NIL THEN
    RETURN NIL;
  END;
  e.CopyMem(from,new^,len);
  RETURN new;
END CopyStringAdd;


PROCEDURE CopyCString * (from: e.STRPTR): BT.DynString;
BEGIN
  RETURN CopyCStringAdd(from,0);
END CopyCString;


PROCEDURE CopyString * (from: ARRAY OF CHAR): BT.DynString;
(* $CopyArrays- *)
BEGIN
  RETURN CopyStringAdd(from,0);
END CopyString;

(*
 * searches for occurance of charakter
 *)
PROCEDURE OccursChar * (VAR s: ARRAY OF CHAR;
                            c: CHAR): LONGINT;
VAR
  i: LONGINT;
BEGIN
  i := 0;
  WHILE (i <= LEN(s)) AND (s[i] # 0X) DO
    IF s[i] = c THEN
      RETURN i; END;
    INC(i);
  END;
  RETURN -1;
END OccursChar;


(*
 * searches for occurance of charakter starting at <pos>
 * if <pos> is negative, searches backwards starting form <pos>
 *)
PROCEDURE OccursCharPos*(VAR s: ARRAY OF CHAR;
                             c: CHAR;
                         start: LONGINT): LONGINT;
VAR
  i: LONGINT;
BEGIN
  IF start < 0 THEN
    start := -start;
    i := str.Length(s)-1;
    IF start >= i THEN start := i; END;
    WHILE start >= 0 DO
      IF s[start] = c THEN
        RETURN start; END;
      DEC(start);
    END;
  ELSE
    WHILE (start <= LEN(s)) & (s[start] # 0X) DO
      IF s[start] = c THEN
        RETURN start; END;
      INC(start);
    END;
  END;
  RETURN -1;
END OccursCharPos;


PROCEDURE StrCmpN*(s1, s2: ARRAY OF CHAR;      (* $CopyArrays- *)
                   maxlen: LONGINT): INTEGER;
VAR
  Diff: INTEGER;
  i: LONGINT;
  Res, Max: LONGINT;
BEGIN
  Res := LEN(s1)-LEN(s2);
  IF Res < 0 THEN
    Max := LEN(s1);
  ELSE
    Max := LEN(s2);
  END;
  IF Max > maxlen THEN
    Max := maxlen; END;

  i := 0;
  WHILE i < Max DO
    Diff := ORD(s1[i])-ORD(s2[i]);
    IF (Diff # 0) OR (s1[i] = 0X) THEN
      RETURN Diff
    END;
    INC(i);
  END;

  IF i = maxlen THEN
    RETURN 0; END;

  (* LEN(s1) = Length(s2)? und umgekehrt *)
  IF (Res < 0) AND (s2[i] # 0X) THEN
    RETURN -1;
  ELSIF (Res > 0) AND (s1[i] # 0X) THEN
    RETURN 1;
  END;
  RETURN 0;
END StrCmpN;


PROCEDURE NCStrCmpN*(s1, s2: ARRAY OF CHAR;     (* $CopyArrays- *)
                     maxlen: LONGINT): INTEGER;
VAR
  Diff: INTEGER;
  i: LONGINT;
  Res, Max: LONGINT;
BEGIN
  Res := LEN(s1)-LEN(s2);
  IF Res < 0 THEN
    Max := LEN(s1);
  ELSE
    Max := LEN(s2);
  END;
  IF Max > maxlen THEN
    Max := maxlen; END;

  i := 0;
  WHILE i < Max DO
    Diff := ORD(CAP(s1[i]))-ORD(CAP(s2[i]));
    IF (Diff # 0) OR (s1[i] = 0X) THEN
      RETURN Diff
    END;
    INC(i);
  END;

  IF i = maxlen THEN
    RETURN 0; END;

  (* LEN(s1) = Length(s2)? und umgekehrt *)
  IF (Res < 0) AND (s2[i] # 0X) THEN
    RETURN -1;
  ELSIF (Res > 0) AND (s1[i] # 0X) THEN
    RETURN 1;
  END;
  RETURN 0;
END NCStrCmpN;


PROCEDURE NCStrCmp*(s1, s2: ARRAY OF CHAR): INTEGER; (* $CopyArrays- *)
VAR
  Diff: INTEGER;
  i: LONGINT;
  Res, Max: LONGINT;
BEGIN
  Res := LEN(s1)-LEN(s2);
  IF Res < 0 THEN
    Max := LEN(s1);
  ELSE
    Max := LEN(s2);
  END;

  i := 0;
  WHILE i < Max DO
    Diff := ORD(CAP(s1[i]))-ORD(CAP(s2[i]));
    IF (Diff#0) OR (s1[i] = 0X) THEN
      RETURN Diff
    END;
    INC(i);
  END;

  (* LEN(s1) = Length(s2)? und umgekehrt *)
  IF (Res < 0) AND (s2[i] # 0X) THEN
    RETURN -1;
  ELSIF (Res > 0) AND (s1[i] # 0X) THEN
    RETURN 1;
  END;
  RETURN 0;
END NCStrCmp;


PROCEDURE WordLen*(s: ARRAY OF CHAR; start: LONGINT): LONGINT;
VAR
  i: LONGINT;
BEGIN
  i := 0;
  WHILE start < LEN(s) DO
    CASE s[start] OF
      CHR(0), " ": RETURN i;
    ELSE END;
    INC(i); INC(start);
  END;
  RETURN i;
END WordLen;

(*
 * n-ter Buchstabe ist LastNospace, d.h.
 * (s[n-1] # " ") & ((s[n] = " ") OR (s[n] = 0X))
 *)
PROCEDURE LastNoSpace*(s: ARRAY OF CHAR): LONGINT;
VAR
  i: LONGINT;
BEGIN
  i := str.Length(s);
  REPEAT DEC(i); UNTIL (i < 0) OR (s[i] # " ");
  RETURN i+1;
END LastNoSpace;

(*
 * returns index of first no space char or -1, if string empty
 *)
PROCEDURE FirstNoSpace*(s: ARRAY OF CHAR): LONGINT;
VAR
  i: LONGINT;
BEGIN
  i := 0;
  WHILE (i < LEN(s)) DO
    CASE s[i] OF
      " ": INC(i);
    | CHR(0): RETURN -1; (* empty *)
    ELSE
      RETURN i (* no space char *)
    END;
  END;
  RETURN -1;
END FirstNoSpace;

(*
 * shifts String right; if <shift> < 0 to left
 *)
PROCEDURE Shift*(VAR string: ARRAY OF CHAR; shift: LONGINT);
BEGIN
  IF shift < 0 THEN
    bmove(y.ADR(string[-shift]),y.ADR(string),LEN(string)+shift);
    string[LEN(string)+shift] := CHR(0);
  ELSE
    bmove(y.ADR(string),y.ADR(string[shift]),LEN(string)-shift);
    bfill(y.ADR(string), ' ', shift);
  END;
END Shift;


PROCEDURE SuperWrite*(VAR string: ARRAY OF CHAR;
                         overlay: ARRAY OF CHAR;
                             pos: LONGINT;
                      start, len: LONGINT);
VAR
  k, l: LONGINT;
BEGIN
  k := str.Length(string);
  l := str.Length(overlay);

  IF (len < 0)  THEN len := l; END;

  IF start + len > l THEN
    len := l-start; END;
  IF pos + len > LEN(string) THEN
    len := LEN(string)-pos; END;
  WHILE len > 0 DO
    string[pos] := overlay[start];
    INC(pos); INC(start);
    DEC(len);
  END;
  IF (pos > k) & (pos < LEN(string)) THEN string[pos] := CHR(0); END;
END SuperWrite;


PROCEDURE StripSpaces * (VAR in: ARRAY OF CHAR): LONGINT;
VAR
  to, from: LONGINT;
  lastWasSpace: BOOLEAN;
BEGIN
  from := 0; to := 0; lastWasSpace := TRUE; (* remove leading spaces *)
  LOOP
    IF from >= LEN(in) THEN EXIT; END;
    CASE in[from] OF
    | CHR(0): in[to] := CHR(0); EXIT;
    | " ": IF ~ lastWasSpace THEN in[to] := in[from]; INC(to); END;
           lastWasSpace := TRUE; (* copy one space *)
    ELSE
      in[to] := in[from]; INC(to); lastWasSpace := FALSE;
    END;
    INC(from);
  END;
  RETURN to; (* length of string *)
END StripSpaces;

(*
 * KillBlanks : lscht Blanks aus einem String. Ist ntig, um nach
 *              Conversion.IntToString () den String zu bereinigen
 *)
PROCEDURE KillSpaces * (VAR in: ARRAY OF CHAR): LONGINT;
VAR
  to, from: LONGINT;
BEGIN
  from := 0; to := 0;
  LOOP
    IF from >= LEN(in) THEN EXIT; END;
    CASE in[from] OF
    | CHR(0): in[to] := CHR(0); EXIT;
    | " ": (* do nothing, just ignore *)
    ELSE
      in[to] := in[from]; INC(to);
    END;
    INC(from);
  END;
  RETURN to; (* length of string *)
END KillSpaces;


PROCEDURE JustifyFull*(VAR string: ARRAY OF CHAR; margin: LONGINT);
VAR
  i: LONGINT;
  firstnb, lastnb, n, fill, sp: LONGINT;
  s: e.STRPTR;
  c: CHAR;
BEGIN
  IF margin > LEN(string) THEN margin := LEN(string); END;
  firstnb := FirstNoSpace(string);
  lastnb := LastNoSpace(string)-1;
  IF firstnb >= lastnb THEN RETURN END;

  fill := margin - lastnb - 1;
  IF fill <= 0 THEN RETURN END;

  (** lastnb < LEN(string) **)
  string[lastnb + 1] := CHR(0); (* strip string *)
  n := 0;
  i := firstnb;
  WHILE i <= lastnb DO
    LOOP
      CASE string[i] OF
        CHR(0), CHR(32): EXIT;
      ELSE
        INC(i);
      END;
    END;
    IF (i <= lastnb) THEN
      INC(n);                        (* Anzahl Lcken *)
      WHILE (string[i] = 20X) DO INC(i); END;
    END;
  END;
  i := firstnb;
  WHILE (n > 0) & (fill > 0) & (string[i] # CHR(0)) DO
    LOOP
      CASE string[i] OF
        CHR(0), CHR(32): EXIT;
      ELSE
        INC(i);
      END;
    END;
    sp := fill DIV n; (* number of spaces to fill in *)
    s := y.ADR(string[i]);
    bmove(s,y.ADR(s[sp]),CLength(s)+1);
    bfill(s, ' ', sp);
    WHILE (string[i] = 20X) DO
      INC(i); END;
    DEC(fill,sp);
    DEC(n);
  END;
END JustifyFull;

(*
 * ChangeBlanks () : Blanks, / und : durch c ersetzen, damit Strings sofort
 *                   und ohne weitere berprfungen als Filenamen verwendet
 *                   werden knnen.
 *)
PROCEDURE ChangeBlanks * (VAR in : ARRAY OF CHAR; c: CHAR);
VAR
  to: LONGINT;
BEGIN
  to := 0;
  LOOP
    IF to >= LEN(in) THEN EXIT; END;
    CASE in[to] OF
    | CHR(0): in[to] := CHR(0); EXIT;
    | " ", "/", ":": in[to] := c;
    ELSE END;
    INC(to);
  END;
END ChangeBlanks;

(*
(* ANSI-Steuerzeichen loeschen        *)
PROCEDURE StripANSI (VAR string: ARRAY OF CHAR): LONGINT;
VAR
  i, Count: LONGINT;
  inESC, inCSI: BOOLEAN;
BEGIN
  i := 0; count := 0; inESC := FALSE; inCSI := FALSE;

  LOOP
    IF i = LEN(string) THEN EXIT; END;
    CASE string[i] OF
    CHR(0): EXIT;      (* end of string *)

    asc.can, asc.sub
        (* Wenn ein CAN (ASCII 24) oder SUB (ASCII 26) auftritt, wird die    *)
        (* aktuelle ANSI-Sequenz sofort abgebrochen                          *)
      inESC := FALSE;
      inCSI := FALSE;
    | asc.esc:
        inESC := TRUE;  inCSI := FALSE; (* CSI-Sequenzen werden abgebrochen  *)
    | asc.csi:
        inESC := FALSE; inCSI := TRUE;  (* ESC-Sequenzen werden abgebrochen  *)
    ELSE
      IF inESC = TRUE THEN             (* Ist es eine ESC-Sequenz ?         *)

        IF string[i] = "[" THEN        (* 7-Bit-CSI-Sequenz ?               *)
          inCSI := TRUE;               (* CSI-Sequenz ein                   *)
          inESC := FALSE;              (* ESC-Sequenz aus                   *)

        ELSIF string[i] >= "0" THEN
                    (* ESC-Sequenzen bestehen nur aus einem `ESC' und          *)
                    (* einer ASCII-Zahl, sobald ein Zeichen groesser ASCII `0' *)
                    (* auftritt, ist die ESC-Sequenz beendet.                  *)
          inESC := FALSE;
        END;

      ELSIF inCSI = TRUE THEN          (* Ist es ein CSI-Kommando ?         *)

        IF string[i] >= "@" THEN
                    (* Sobald in einer CSI-Sequenz ein Zeichen groesser oder   *)
                    (* gleich ASCII `@' auftritt, ist die CSI-Sequenz beendet  *)
          inCSI := FALSE;
        END;

      ELSIF (inESC = FALSE) AND (inCSI = FALSE) THEN

        string[count] := string[i];    (* Zeichen in Zielstring kopieren    *)
        INC(count);                    (* und Zaehler erhoehen                *)
      END;
    END;
    INC(i);
  END;                         (* Ende der Such- und Kopierschleife  *)
  IF count < LEN(string) THEN
    string[count] := CHR(0);
  END;
  RETURN count;                (* Zielstringlaenge zurueckgeben      *)
END StripANSI;
*)
END MoreStrings.

